from openai_load_balancer.load_balancer import LoadBalancer


class OpenAILoadBalancer:
    class ChatCompletion:
        def __init__(self, load_balancer: LoadBalancer):
            self.load_balancer = load_balancer

        def create(self, **kwargs):
            return self.load_balancer.try_send_request('chat_completion_create', **kwargs)

    class Completion:
        def __init__(self, load_balancer: LoadBalancer):
            self.load_balancer = load_balancer

        def create(self, **kwargs):
            return self.load_balancer.try_send_request('completion_create', **kwargs)

    class Embedding:
        def __init__(self, load_balancer: LoadBalancer):
            self.load_balancer = load_balancer

        def create(self, **kwargs):
            return self.load_balancer.try_send_request('embedding_create', **kwargs)

    def __init__(self, load_balancer: LoadBalancer):
        self.load_balancer = load_balancer
        self.ChatCompletion = OpenAILoadBalancer.ChatCompletion(load_balancer)
        self.Completion = OpenAILoadBalancer.Completion(load_balancer)
        self.Embedding = OpenAILoadBalancer.Embedding(load_balancer)



from datetime import timedelta
import pytest
from unittest.mock import patch, create_autospec, Mock
from openai_load_balancer.load_balancer import LoadBalancer
from openai_load_balancer.api_endpoint import ApiEndpoint


@pytest.fixture
def load_balancer():
    endpoint_configs = [
        {
            "api_type": "azure",
            "base_url": "AZURE_API_BASE_URL_1",
            "api_key_env": "AZURE_API_KEY_1",
            "version": "2023-05-15"
        },
        {
            "api_type": "open_ai",
            "base_url": "https://api.openai.com/v1",
            "api_key_env": "OPENAI_API_KEY_1",
            "version": None
        },
    ]
    FAILURE_THRESHOLD = 5
    COOLDOWN_PERIOD = timedelta(minutes=10)

    return LoadBalancer(endpoint_configs, failure_threshold=FAILURE_THRESHOLD, cooldown_period=COOLDOWN_PERIOD)


def test_round_robin(load_balancer):
    first_endpoint = load_balancer.get_next_active_endpoint()
    second_endpoint = load_balancer.get_next_active_endpoint()
    assert first_endpoint != second_endpoint


def test_100_round_robin(load_balancer):
    # create 100 mock endpoints
    mock_endpoints = [create_autospec(
        ApiEndpoint, instance=True) for _ in range(100)]
    # configure all mock endpoints to be active
    for mock_endpoint in mock_endpoints:
        mock_endpoint.is_active.return_value = True
    # replace actual endpoints with mock endpoints
    load_balancer.api_endpoints = mock_endpoints
    # test getting the next active endpoint 100 times, and assert that it's the next endpoint as we expect
    for i in range(100):
        next_endpoint = load_balancer.get_next_active_endpoint()
        assert next_endpoint == mock_endpoints[i]


@patch('openai_load_balancer.load_balancer.LoadBalancer.send_request')
def test_handle_endpoint_failure(mock_send_request, load_balancer):
    # Create separate mock instances for each endpoint
    mock_endpoint1 = create_autospec(ApiEndpoint, instance=True)
    mock_endpoint2 = create_autospec(ApiEndpoint, instance=True)
    mock_endpoint3 = create_autospec(ApiEndpoint, instance=True)

    # Configure the first endpoint to simulate a failure
    mock_endpoint1.is_active.return_value = True
    mock_endpoint1.mark_failed = Mock()

    # Configure the second endpoint to simulate success
    mock_endpoint2.is_active.return_value = True
    mock_endpoint2.mark_failed = Mock()

    # Configure the third endpoint (should not be used in this test)
    mock_endpoint3.is_active.return_value = True
    mock_endpoint3.mark_failed = Mock()

    # Replace actual endpoints with mocks
    load_balancer.api_endpoints = [
        mock_endpoint1, mock_endpoint2, mock_endpoint3]

    # Setup mock for send_request
    mock_send_request.side_effect = [Exception("Failed request"), "Success"]

    # Make a request - it should fail on the first endpoint and succeed on the second
    result = load_balancer.try_send_request('test_method', test_arg='value')

    # Assert that the send_request was called twice (once for each endpoint)
    assert mock_send_request.call_count == 2
    assert result == "Success"
    assert mock_endpoint1.mark_failed.called
    assert not mock_endpoint3.mark_failed.called


@patch('openai_load_balancer.load_balancer.LoadBalancer.send_request')
def test_all_endpoints_failure(mock_send_request, load_balancer):
    # Create a list of mock endpoints
    mock_endpoints = [create_autospec(ApiEndpoint, instance=True) for _ in range(
        len(load_balancer.api_endpoints))]

    # Configure all mock endpoints to simulate failure
    for mock_endpoint in mock_endpoints:
        mock_endpoint.is_active.return_value = True
        mock_endpoint.mark_failed = Mock()

    # Replace actual endpoints with mock endpoints
    load_balancer.api_endpoints = mock_endpoints

    # Setup mock for send_request to always raise an exception
    mock_send_request.side_effect = Exception("Failed request")

    # Attempt to make a request, expecting it to raise an exception
    with pytest.raises(Exception) as excinfo:
        load_balancer.try_send_request('test_method', test_arg='value')

    # Assert that the exception message is as expected
    assert str(excinfo.value) == "All endpoints failed."

    # Assert that mark_failed was called on each endpoint
    for mock_endpoint in mock_endpoints:
        assert mock_endpoint.mark_failed.called


def test_load_balancing_disabled(load_balancer):
    """Test that when load balancing is disabled, the LoadBalancer always tries the first endpoint unless it's inactive."""
    load_balancer.load_balancing_enabled = False
    first_endpoint = load_balancer.get_next_active_endpoint()

    # test getting the next active endpoint 10 times
    for _ in range(10):
        next_endpoint = load_balancer.get_next_active_endpoint()
        assert next_endpoint == first_endpoint


def test_load_balancing_disabled_tries_next_endpoint(load_balancer):
    """Test that when load balancing is disabled, the LoadBalancer tries the next endpoint if the first one is inactive."""
    # Create mock endpoints
    mock_endpoint1 = create_autospec(ApiEndpoint, instance=True)
    mock_endpoint2 = create_autospec(ApiEndpoint, instance=True)

    # Configure the first mock endpoint to be inactive
    mock_endpoint1.is_active.return_value = False

    # Configure the second mock endpoint to be active
    mock_endpoint2.is_active.return_value = True

    # Replace actual endpoints with mocks
    load_balancer.api_endpoints = [mock_endpoint1, mock_endpoint2]
    # Disable load balancing
    load_balancer.load_balancing_enabled = False

    # Get the next active endpoint
    next_active_endpoint = load_balancer.get_next_active_endpoint()

    # Assert that the second endpoint is returned, not the first
    assert next_active_endpoint == mock_endpoint2


def test_skipping_inactive_endpoints(load_balancer):
    # Create mock endpoints
    mock_endpoint1 = create_autospec(ApiEndpoint, instance=True)
    mock_endpoint2 = create_autospec(ApiEndpoint, instance=True)
    mock_endpoint1.is_active.return_value = False
    mock_endpoint2.is_active.return_value = True

    load_balancer.api_endpoints = [mock_endpoint1, mock_endpoint2]

    # The load balancer should skip the first inactive endpoint
    assert load_balancer.get_next_active_endpoint() == mock_endpoint2


@patch('openai_load_balancer.load_balancer.LoadBalancer.send_request')
def test_send_request_arguments(mock_send_request, load_balancer):
    """Test that the correct arguments are passed to the send_request method."""
    mock_endpoint = create_autospec(ApiEndpoint, instance=True)
    mock_endpoint.is_active.return_value = True
    load_balancer.api_endpoints = [mock_endpoint]

    test_args = {'arg1': 'value1', 'arg2': 'value2'}
    load_balancer.try_send_request('test_method', **test_args)

    mock_send_request.assert_called_with(
        mock_endpoint, 'test_method', **test_args)


@patch('openai_load_balancer.load_balancer.LoadBalancer.send_request', return_value="Success")
def test_failure_count_reset_on_success(mock_send_request, load_balancer):
    mock_endpoint = create_autospec(ApiEndpoint, instance=True)
    mock_endpoint.is_active.return_value = True
    load_balancer.api_endpoints = [mock_endpoint]

    load_balancer.try_send_request('test_method', test_arg='value')
    mock_endpoint.reset.assert_called()


def test_all_endpoints_inactive(load_balancer):
    mock_endpoints = [create_autospec(ApiEndpoint, instance=True) for _ in range(
        len(load_balancer.api_endpoints))]
    for mock_endpoint in mock_endpoints:
        mock_endpoint.is_active.return_value = False
        mock_endpoint.mark_failed = Mock()
    load_balancer.api_endpoints = mock_endpoints
    with pytest.raises(Exception) as excinfo:
        load_balancer.try_send_request('test_method', test_arg='value')
    assert str(excinfo.value) == "All endpoints are inactive."



import pytest
from unittest.mock import Mock
from openai_load_balancer.openai_interface import OpenAILoadBalancer


@pytest.fixture
def mock_load_balancer():
    mock = Mock()
    return mock


@pytest.fixture
def openai_load_balancer(mock_load_balancer):
    return OpenAILoadBalancer(mock_load_balancer)


def test_chat_completion_create(openai_load_balancer, mock_load_balancer):
    test_kwargs = {"messages": [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "Hello!"}
    ], "model": "gpt-3.5-turbo"}
    openai_load_balancer.ChatCompletion.create(**test_kwargs)
    mock_load_balancer.try_send_request.assert_called_once_with(
        'chat_completion_create', **test_kwargs)


def test_completion_create(openai_load_balancer, mock_load_balancer):
    test_kwargs = {"prompt": "Hello", "model": "text-davinci-003"}
    openai_load_balancer.Completion.create(**test_kwargs)
    mock_load_balancer.try_send_request.assert_called_once_with(
        'completion_create', **test_kwargs)


def test_embedding_create(openai_load_balancer, mock_load_balancer):
    test_kwargs = {"input": "Hello", "model": "text-embedding-ada-002"}
    openai_load_balancer.Embedding.create(**test_kwargs)
    mock_load_balancer.try_send_request.assert_called_once_with(
        'embedding_create', **test_kwargs)



from datetime import datetime, timedelta
import pytest
from openai_load_balancer.api_endpoint import ApiEndpoint


@pytest.fixture
def api_endpoint():
    return ApiEndpoint(api_type="open_ai", base_url="https://api.openai.com/v1", api_key_env="OPENAI_API_KEY")


def test_api_endpoint_initial_state(api_endpoint):
    assert api_endpoint.failure_count == 0
    assert api_endpoint.last_failed_time is None


def test_api_endpoint_is_active(api_endpoint):
    assert api_endpoint.is_active(
        failure_threshold=5, cooldown_period=timedelta(minutes=10))


def test_api_endpoint_mark_failed(api_endpoint):
    api_endpoint.mark_failed()
    assert api_endpoint.failure_count == 1
    assert api_endpoint.last_failed_time is not None


def test_api_endpoint_reset(api_endpoint):
    api_endpoint.mark_failed()
    api_endpoint.reset()
    assert api_endpoint.failure_count == 0
    assert api_endpoint.last_failed_time is None


def test_api_endpoint_reactivate_after_cooldown(api_endpoint):
    api_endpoint.mark_failed()
    api_endpoint.last_failed_time = datetime.now() - timedelta(minutes=11)
    assert api_endpoint.is_active(
        failure_threshold=5, cooldown_period=timedelta(minutes=10))


def test_api_endpoint_active_until_failure_threshold(api_endpoint):
    for _ in range(4):
        api_endpoint.mark_failed()
    assert api_endpoint.is_active(
        failure_threshold=5, cooldown_period=timedelta(minutes=10))


def test_api_endpoint_inactive_after_failure_threshold(api_endpoint):
    for _ in range(5):
        api_endpoint.mark_failed()
    assert not api_endpoint.is_active(
        failure_threshold=5, cooldown_period=timedelta(minutes=10))


def test_api_endpoint_inactive_before_cooldown(api_endpoint):
    # Increment failure_count to reach the threshold
    for _ in range(5):  # Assuming failure_threshold is 5
        api_endpoint.mark_failed()

    # Set last_failed_time to 5 minutes ago (inside the cooldown period)
    api_endpoint.last_failed_time = datetime.now() - timedelta(minutes=5)

    # Test should now correctly expect the endpoint to be inactive
    assert not api_endpoint.is_active(
        failure_threshold=5, cooldown_period=timedelta(minutes=10))


def test_api_endpoint_failure_count_increment(api_endpoint):
    for i in range(1, 4):  # Test multiple failures
        api_endpoint.mark_failed()
        assert api_endpoint.failure_count == i



import os
import dotenv
import openai
from openai_load_balancer.api_endpoint import ApiEndpoint
from tenacity import retry, wait_random_exponential, stop_after_attempt
import threading
dotenv.load_dotenv()


class LoadBalancer:
    def __init__(self, endpoint_configs, failure_threshold, cooldown_period, load_balancing_enabled=True, model_engine_mapping=None):
        """Initializes the load balancer with the passed in endpoint configurations and other configs"""
        self.api_endpoints = [ApiEndpoint(**config)
                              for config in endpoint_configs]
        self.current_index = 0
        self.failure_threshold = failure_threshold
        self.cooldown_period = cooldown_period
        self.load_balancing_enabled = load_balancing_enabled
        self.model_engine_mapping = model_engine_mapping
        self.lock = threading.Lock()  # Lock for thread safety

    def get_next_active_endpoint(self):
        """Gets the next active endpoint to use. If load balancing is disabled, always returns the first endpoint, unless the first endpoint is in_active, then proceeds to find the next one. If load balancing is enabled, returns the next active endpoint."""
        with self.lock:  # Acquire lock for thread-safe access
            if not self.load_balancing_enabled:
                # If load balancing is disabled, always try the first active endpoint
                if self.api_endpoints[0].is_active(self.failure_threshold, self.cooldown_period):
                    return self.api_endpoints[0]
                else:
                    # If the first endpoint is inactive, proceed to find the next active one
                    self.current_index = 1

            for _ in range(len(self.api_endpoints)):
                # get the current endpoint
                endpoint = self.api_endpoints[self.current_index]

                # Increment the current index (so the next time we call this function, we'll get the next endpoint in the list). If we've reached the end of the list, loop back to the beginning
                self.current_index = (
                    self.current_index + 1) % len(self.api_endpoints)

                if endpoint.is_active(self.failure_threshold, self.cooldown_period):
                    return endpoint

            # If we've tried all endpoints and none are active, raise an exception
            raise Exception("All endpoints are inactive.")

    @retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(5))
    def send_request(self, endpoint, method_name, **kwargs):
        """Calls OpenAI's API with the corresponding method and arguments to the passed in endpoint. If it fails, raises an exception"""
        # openai has a standard base_url, whereas for azure we'll read it from the environment variable
        openai.api_base = str(os.getenv(
            endpoint.base_url)) if endpoint.api_type == "azure" else endpoint.base_url
        openai.api_version = endpoint.version
        openai.api_key = str(os.getenv(endpoint.api_key_env))
        openai.api_type = endpoint.api_type

        model_engine_mapping = self.model_engine_mapping or {}

        # Adjust arguments for Azure
        if endpoint.api_type == "azure":
            # In Azure, instead of using the model keyword, you use the engine keyword. Get the appropriate engine name for the passed in model
            if "model" in kwargs:
                engine_name = model_engine_mapping.get(
                    kwargs["model"], kwargs["model"])
                kwargs["engine"] = engine_name
                del kwargs["model"]
        if endpoint.api_type == "open_ai":
            # Do the same for switching from Azure engine to OpenAI model
            if "engine" in kwargs:
                # since MODEL_ENGINE_MAPPING is a dict going from model -> engine, we need to invert it to get engine -> model
                model_name = {v: k for k, v in model_engine_mapping.items()}.get(
                    kwargs["engine"], kwargs["engine"])
                kwargs["model"] = model_name
                del kwargs["engine"]
        # Map method_name to the actual OpenAI function
        if method_name == 'completion_create':
            response = openai.Completion.create(**kwargs)
        elif method_name == 'chat_completion_create':
            response = openai.ChatCompletion.create(**kwargs)
        elif method_name == 'embedding_create':
            response = openai.Embedding.create(**kwargs)

        endpoint.failure_count = 0  # Reset on successful request
        return response

    def try_send_request(self, method_name, **kwargs):
        """Try to send the request to active endpoints. If it fails, mark as failed and try the next active endpoint."""
        for _ in range(len(self.api_endpoints)):
            endpoint = self.get_next_active_endpoint()

            try:
                response = self.send_request(endpoint, method_name, **kwargs)
                # Reset the endpoint on a successful request
                endpoint.reset()
                return response
            except Exception as e:
                # Mark the endpoint as failed
                endpoint.mark_failed()

        # If all endpoints have been tried and failed, raise an exception
        raise Exception("All endpoints failed.")



from .load_balancer import LoadBalancer
from .openai_interface import OpenAILoadBalancer
from datetime import timedelta

DEFAULT_MODEL_ENGINE_MAPPING = {
    "gpt-4": "gpt4",
    "gpt-3.5-turbo": "gpt-35-turbo",
    "text-embedding-ada-002": "text-embedding-ada-002"
}


def initialize_load_balancer(endpoints, model_engine_mapping=DEFAULT_MODEL_ENGINE_MAPPING, failure_threshold=5, cooldown_period=timedelta(minutes=10), load_balancing_enabled=True):
    """Initializes the load balancer with the endpoint settings and other configs. 
    @param endpoints: A list of dictionaries containing the OpenAI API endpoint configurations. 
    @param model_engine_mapping: A dictionary mapping the OpenAI model names to the Azure engine names.
    @param failure_threshold: The number of consecutive failures of a request to an endpoint before the endpoint is temporarily marked as inactive
    @param cooldown_period: The minimum amount of time an endpoint is marked as inactive before it is reset to active.
    @param load_balancing_enabled: Whether or not to enable load balancing. If false, the first active endpoint will always be used, and other endpoints will only be used in case the first one fails.
    """
    load_balancer = LoadBalancer(
        endpoints,
        failure_threshold=failure_threshold,
        cooldown_period=cooldown_period,
        load_balancing_enabled=load_balancing_enabled, model_engine_mapping=model_engine_mapping
    )
    return OpenAILoadBalancer(load_balancer)



from setuptools import setup, find_packages

setup(
    name='openai-load-balancer',
    version='0.1.1',
    packages=find_packages(),
    include_package_data=True,
    license='MIT',
    description='The OpenAI Load Balancer is a Python library designed to distribute API requests across multiple endpoints (supports both OpenAI and Azure OpenAI). It implements a round-robin mechanism for load balancing and includes exponential backoff for retrying failed requests.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Rick Liu',
    author_email='rick@spryngtime.com',
    url='https://github.com/Spryngtime/openai-load-balancer',
    install_requires=[
        # Any required packages here
        'openai',
        'python-dotenv',
        'tenacity'
    ],
    classifiers=[
        # Full list at https://pypi.org/classifiers/
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        # Add additional supported Python versions
    ],
)



import threading
from datetime import datetime


class ApiEndpoint:
    def __init__(self, api_type, base_url, api_key_env, version=None):
        """Inits an API endpoint based on the passed in configuration. You can make adjustments to your configurations in config.py"""
        self.api_type = api_type
        self.base_url = base_url
        self.api_key_env = api_key_env
        self.version = version
        self.failure_count = 0
        self.last_failed_time = None
        self.lock = threading.Lock()  # Adding a lock for thread safety

    def is_active(self, failure_threshold, cooldown_period):
        """Checks if the endpoint is active. If it has failed more than failure_threshold times, it is inactive. If it is currently marked as failed with a last_failed_time within the cooldown_period, it is inactive. If the cooldown_period has passed since the last failure, the endpoint will be reset to active and returns true."""
        with self.lock:  # Ensure thread-safe read
            if self.failure_count < failure_threshold:
                return True
            if self.last_failed_time and datetime.now() - self.last_failed_time > cooldown_period:
                self.reset()
                return True
            return False

    def reset(self):
        """Resets the endpoint to active by setting failure_count to 0 and last_failed_time to None"""
        with self.lock:  # Ensure thread-safe state update
            self.failure_count = 0
            self.last_failed_time = None

    def mark_failed(self):
        """Marks the endpoint as failed by incrementing failure_count and setting last_failed_time to the current time"""
        with self.lock:  # Ensure thread-safe state update
            self.failure_count += 1
            self.last_failed_time = datetime.now()
